from asyncio import Semaphore, create_task, sleep
from typing import Protocol


class GatewayRateLimiter(Protocol):
    per: int

    def __init__(self, rate: int, per: int) -> None:
        ...

    async def wait(self) -> None:
        ...


class LocalGatewayRateLimiter:
    def __init__(self, rate: int, per: int) -> None:
        self.per = per

        self._lock = Semaphore(rate)

    async def _release(self, after: float) -> None:
        await sleep(after)

        self._lock.release()

    async def wait(self) -> None:
        await self._lock.acquire()

        create_task(self._release(self.per))



from __future__ import annotations

from asyncio import create_task, sleep
from collections import defaultdict
from dataclasses import dataclass
from json import dumps
from typing import Any, Awaitable, Callable, Mapping, Optional, Sequence, Type, Union

from aiohttp import BasicAuth, ClientResponse, ClientSession, FormData

from bauxite.constants import API_URL, VERSION

from .errors import (
    BadGateway,
    BadRequest,
    Forbidden,
    GatewayTimeout,
    HTTPError,
    MethodNotAllowed,
    NotFound,
    ServerError,
    ServiceUnavailable,
    TooManyRequests,
    Unauthorized,
    UnprocessableEntity,
)
from .file import File
from .ratelimiting import LocalRateLimiter, RateLimiter
from .route import Route

Callback = Callable[[ClientResponse, Route], Awaitable[None]]
Unset = object()


@dataclass
class _RequestContext:
    route: Route
    headers: dict[str, str]
    params: dict[str, Any]
    files: Sequence[File]
    json: Any


@dataclass
class _ResponseContext:
    route: Route
    response: ClientResponse
    successful: bool


class HTTPClient:
    _status_codes: Mapping[int, Type[HTTPError]] = defaultdict(
        lambda: HTTPError,
        {
            400: BadRequest,
            401: Unauthorized,
            403: Forbidden,
            404: NotFound,
            405: MethodNotAllowed,
            422: UnprocessableEntity,
            429: TooManyRequests,
            500: ServerError,
            502: BadGateway,
            503: ServiceUnavailable,
            504: GatewayTimeout,
        },
    )

    def __init__(
        self,
        token: str,
        api_url: Optional[str] = None,
        user_agent: Optional[str] = None,
        proxy_url: Optional[str] = None,
        proxy_auth: Optional[BasicAuth] = None,
        ratelimiter: Optional[RateLimiter] = None,
        on_success: Optional[set[Callback]] = None,
        on_error: Optional[set[Callback]] = None,
        on_ratelimit: Optional[set[Callback]] = None,
    ) -> None:
        self._token = token.strip()
        self._api_url = api_url or API_URL
        self._user_agent = (
            user_agent
            or f"DiscordBot (https://github.com/vcokltfre/bauxite, {VERSION})"
        )
        self._proxy_url = proxy_url
        self._proxy_auth = proxy_auth
        self._ratelimiter = ratelimiter or LocalRateLimiter()

        self.__session: Optional[ClientSession] = None

        self._on_success = on_success or set()
        self._on_error = on_error or set()
        self._on_ratelimit = on_ratelimit or set()

    @property
    def _session(self) -> ClientSession:
        if self.__session and not self.__session.closed:
            return self.__session

        self.__session = ClientSession(
            headers={
                "Authorization": f"Bot {self._token}",
                "User-Agent": self._user_agent,
            }
        )

        return self.__session

    def _dispatch(self, listeners: set[Callback], ctx: _ResponseContext) -> None:
        for listener in listeners:
            create_task(listener(ctx.response, ctx.route))

    async def _request(
        self, ctx: _RequestContext, reset_files: int
    ) -> _ResponseContext:
        if ctx.files:
            data = FormData()

            for i, file in enumerate(ctx.files):
                file.reset(reset_files)

                data.add_field(f"file_{i}", file.fp, filename=file.filename)

            if ctx.json is not Unset:
                data.add_field(
                    "payload_json", dumps(ctx.json), content_type="application/json"
                )

            ctx.params["data"] = data
        elif ctx.json is not Unset:
            ctx.params["json"] = ctx.json

        lock = await self._ratelimiter.acquire(ctx.route.bucket)

        async with lock:
            response = await self._session.request(
                ctx.route.method,
                self._api_url + ctx.route.path,
                headers=ctx.headers,
                **ctx.params,
            )

            status = response.status
            headers = response.headers

            response_ctx = _ResponseContext(ctx.route, response, 200 <= status < 300)

            rl_reset_after = float(headers.get("X-RateLimit-Reset-After", 0))
            rl_bucket_remaining = int(headers.get("X-RateLimit-Remaining", 1))

            if response_ctx.successful:
                self._dispatch(self._on_success, response_ctx)
                if rl_bucket_remaining == 0:
                    self._dispatch(self._on_ratelimit, response_ctx)
                    await lock.release(rl_reset_after)
                else:
                    await lock.release(0)
                return response_ctx
            elif status == 429:
                self._dispatch(self._on_error, response_ctx)
                self._dispatch(self._on_ratelimit, response_ctx)
                if not headers.get("Via"):
                    raise TooManyRequests(response)

                json = await response.json()

                is_global = json.get("global", False)
                retry_after = json["retry_after"]

                if is_global:
                    await self._ratelimiter.lock_globally(retry_after)
                else:
                    await lock.release(retry_after)
            else:
                self._dispatch(self._on_error, response_ctx)
                raise self._status_codes[status](response)

            return response_ctx

    async def request(
        self,
        route: Route,
        qparams: Optional[dict[str, Union[str, int]]] = None,
        reason: Optional[str] = None,
        files: Optional[Sequence[File]] = None,
        json: Optional[Any] = Unset,
        max_attempts: int = 3,
    ) -> ClientResponse:
        headers = {}
        params = {}

        if qparams:
            params["params"] = qparams

        if reason:
            headers["X-Audit-Log-Reason"] = reason

        for attempt in range(max_attempts):
            ctx = _RequestContext(route, headers, params, files or (), json)
            resp = await self._request(ctx, attempt)

            if resp.successful:
                return resp.response

            if attempt == max_attempts - 1:
                raise self._status_codes[resp.response.status](resp.response)

            await sleep(1 + attempt * 2)

        raise Exception("Unreachable")

    async def close(self) -> None:
        if self.__session:
            await self.__session.close()



from .client import GatewayClient
from .enums import EventDirection, GatewayCloseCodes, GatewayOps, ShardStatus
from .errors import GatewayCriticalError, GatewayReconnect
from .ratelimiting import GatewayRateLimiter, LocalGatewayRateLimiter
from .shard import Shard

__all__ = (
    "EventDirection",
    "GatewayClient",
    "GatewayCloseCodes",
    "GatewayCriticalError",
    "GatewayOps",
    "GatewayRateLimiter",
    "GatewayReconnect",
    "LocalGatewayRateLimiter",
    "Shard",
    "ShardStatus",
)



from .constants import API_URL, VERSION
from .error import BauxiteError
from .gateway import (
    EventDirection,
    GatewayClient,
    GatewayCloseCodes,
    GatewayCriticalError,
    GatewayOps,
    GatewayRateLimiter,
    GatewayReconnect,
    LocalGatewayRateLimiter,
    Shard,
    ShardStatus,
)
from .http import (
    BadGateway,
    BadRequest,
    BucketLock,
    File,
    Forbidden,
    GatewayTimeout,
    HTTPClient,
    HTTPError,
    LocalBucketLock,
    LocalRateLimiter,
    MethodNotAllowed,
    NotFound,
    RateLimiter,
    Route,
    ServerError,
    ServiceUnavailable,
    TooManyRequests,
    Unauthorized,
    UnprocessableEntity,
)

__all__ = (
    "API_URL",
    "VERSION",
    "BadGateway",
    "BadRequest",
    "BauxiteError",
    "BucketLock",
    "Forbidden",
    "GatewayTimeout",
    "HTTPClient",
    "HTTPError",
    "LocalBucketLock",
    "LocalRateLimiter",
    "MethodNotAllowed",
    "NotFound",
    "RateLimiter",
    "Route",
    "ServerError",
    "ServiceUnavailable",
    "TooManyRequests",
    "Unauthorized",
    "UnprocessableEntity",
    "File",
    "EventDirection",
    "GatewayClient",
    "GatewayCloseCodes",
    "GatewayCriticalError",
    "GatewayOps",
    "GatewayRateLimiter",
    "GatewayReconnect",
    "LocalGatewayRateLimiter",
    "Shard",
    "ShardStatus",
)



from asyncio import Task, create_task, sleep
from random import randrange
from sys import platform
from time import time
from typing import Awaitable, Callable, Optional

from aiohttp import (
    ClientSession,
    ClientWebSocketResponse,
    WebSocketError,
    WSMessage,
    WSMsgType,
    WSServerHandshakeError,
)

from .enums import EventDirection, GatewayCloseCodes, GatewayOps, ShardStatus
from .errors import GatewayCriticalError, GatewayReconnect
from .ratelimiting import GatewayRateLimiter, LocalGatewayRateLimiter

CRITICAL = [
    GatewayCloseCodes.NOT_AUTHENTICATED,
    GatewayCloseCodes.AUTHENTICATION_FAILED,
    GatewayCloseCodes.INVALID_API_VERSION,
    GatewayCloseCodes.INVALID_INTENTS,
    GatewayCloseCodes.DISALLOWED_INTENTS,
]

NONCRITICAL = [
    GatewayCloseCodes.INVALID_SEQ,
    GatewayCloseCodes.RATE_LIMITED,
    GatewayCloseCodes.SESSION_TIMEOUT,
]

ShardStatusHook = Callable[["Shard", ShardStatus], Awaitable[None]]


class Shard:
    def __init__(
        self,
        shard_id: int,
        shard_count: int,
        token: str,
        intents: int,
        callback: Callable[["Shard", EventDirection, dict], Awaitable[None]],
        status_hooks: list[ShardStatusHook],
        ratelimiter: Optional[GatewayRateLimiter] = None,
    ) -> None:
        self.id = shard_id

        self._count = shard_count
        self._token = token
        self._intents = intents
        self._callback = callback
        self._hooks = status_hooks
        self._send_limiter = ratelimiter or LocalGatewayRateLimiter(120, 60)

        self._ws: Optional[ClientWebSocketResponse] = None
        self._hb: Optional[float] = None
        self._hb_interval: Optional[float] = None

        self._ack: Optional[bool] = None
        self._last_hb: Optional[float] = None
        self._last_ack: Optional[float] = None

        self._pacemaker: Optional[Task] = None

        self._session: Optional[str] = None
        self._seq: Optional[int] = None

    def __repr__(self) -> str:
        return f"<Shard id={self.id}>"

    @property
    def latency(self) -> Optional[float]:
        if self._last_hb and self._last_ack:
            return self._last_ack - self._last_hb
        return

    def _status_hook(self, status: ShardStatus):
        for hook in self._hooks:
            create_task(hook(self, status))

    async def _spawn_ws(self, session: ClientSession, url: str) -> None:
        self._ws = await session.ws_connect(url)

    async def _connect(self, session: ClientSession, url: str) -> None:
        self._status_hook(ShardStatus.CONNECTING)

        try:
            await self._spawn_ws(session, url)
        except (WebSocketError, WSServerHandshakeError) as e:
            return

        if self._session:
            await self._resume()

        try:
            await self._read()
        except GatewayReconnect:
            pass

    async def connect(self, session: ClientSession, url: str) -> None:
        backoff = 0.01

        while True:
            await sleep(backoff)
            try:
                await self._connect(session, url)
                backoff = 0.01
            except Exception as e:
                if backoff < 5:
                    backoff *= 2

    async def _close(self) -> None:
        self._hb = None

        if self._ws and not self._ws.closed:
            await self._ws.close()

        if self._pacemaker and not self._pacemaker.cancelled():
            self._pacemaker.cancel()

    async def _send(self, message: dict) -> None:
        await self._send_limiter.wait()
        await self._callback(self, EventDirection.OUTBOUND, message)

        try:
            await self._ws.send_json(message)  # type: ignore
        except OSError:
            await self._close()
        except Exception:
            await self._close()
            raise

    async def _identify(self) -> None:
        await self._send(
            {
                "op": GatewayOps.IDENTIFY,
                "d": {
                    "token": self._token,
                    "properties": {
                        "$os": platform,
                        "$browser": "Ablaze",
                        "$device": "Ablaze",
                    },
                    "intents": self._intents,
                    "shard": [self.id, self._count],
                },
            }
        )

    async def _resume(self) -> None:
        self._status_hook(ShardStatus.RESUMING)

        await self._send(
            {
                "op": GatewayOps.RESUME,
                "d": {
                    "token": self._token,
                    "session_id": self._session,
                    "seq": self._seq,
                },
            }
        )

    async def _dispatch(self, data: dict) -> None:
        await self._callback(self, EventDirection.INBOUND, data)

        op = data["op"]

        if op == GatewayOps.HELLO:
            self._pacemaker = create_task(
                self._start_pacemaker(data["d"]["heartbeat_interval"])
            )
            await self._identify()
        elif op == GatewayOps.ACK:
            self._last_ack = time()
            self._ack = True
        elif op == GatewayOps.RECONNECT:
            await self._close()
            raise GatewayReconnect()

    async def _handle_disconnect(self, code: int) -> None:
        """Handle the gateway disconnecting correctly."""

        self._status_hook(ShardStatus.ERRORED)

        if code in CRITICAL:
            raise GatewayCriticalError(code)

        if code in NONCRITICAL:
            self._session = None
            self._seq = None

        await self._close()

        raise GatewayReconnect()

    async def _read(self) -> None:
        self._status_hook(ShardStatus.CONNECTED)

        assert self._ws, "WebSocket is not spawned while _read() is called."

        async for message in self._ws:
            message: WSMessage

            if message.type == WSMsgType.TEXT:
                message_data = message.json()

                if s := message_data.get("s"):
                    self._seq = s

                await self._dispatch(message_data)

        assert self._ws and self._ws.close_code

        await self._handle_disconnect(self._ws.close_code)

    async def _start_pacemaker(self, delay: float) -> None:
        delay = delay / 1000

        await sleep(randrange(0, int(delay)))

        while True:
            if self._last_ack and time() - self._last_ack >= delay:
                return await self._close()

            await self._heartbeat()

            await sleep(delay)

    async def _heartbeat(self) -> None:
        self._last_hb = time()

        await self._send({"op": GatewayOps.HEARTBEAT, "d": self._seq})



from .client import HTTPClient
from .errors import (
    BadGateway,
    BadRequest,
    Forbidden,
    GatewayTimeout,
    HTTPError,
    MethodNotAllowed,
    NotFound,
    ServerError,
    ServiceUnavailable,
    TooManyRequests,
    Unauthorized,
    UnprocessableEntity,
)
from .file import File
from .ratelimiting import BucketLock, LocalBucketLock, LocalRateLimiter, RateLimiter
from .route import Route

__all__ = (
    "BadGateway",
    "BadRequest",
    "BucketLock",
    "File",
    "Forbidden",
    "GatewayTimeout",
    "HTTPClient",
    "HTTPError",
    "LocalBucketLock",
    "LocalRateLimiter",
    "MethodNotAllowed",
    "NotFound",
    "RateLimiter",
    "Route",
    "ServerError",
    "ServiceUnavailable",
    "TooManyRequests",
    "Unauthorized",
    "UnprocessableEntity",
)



from __future__ import annotations

from asyncio import Event, Lock, create_task, sleep
from typing import Protocol


class BucketLock(Protocol):
    async def __aenter__(self) -> "BucketLock":
        ...

    async def __aexit__(self, *args, **kwargs) -> None:
        ...

    async def release(self, after: float = 0) -> None:
        ...


class RateLimiter(Protocol):
    async def acquire(self, bucket: str) -> BucketLock:
        ...

    async def lock_globally(self, release_after: float) -> None:
        ...


class LocalBucketLock:
    def __init__(self) -> None:
        self._lock = Lock()

    async def _release(self, after: float) -> None:
        await sleep(after)
        self._lock.release()

    async def __aenter__(self) -> "LocalBucketLock":
        await self._lock.acquire()
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        pass

    async def release(self, after: float = 0) -> None:
        create_task(self._release(after))


class LocalRateLimiter:
    def __init__(self) -> None:
        self.buckets: dict[str, BucketLock] = {}

        self._global = Event()
        self._global.set()

    async def _lock_global(self, release_after: float) -> None:
        self._global.clear()
        await sleep(release_after)
        self._global.set()

    async def acquire(self, bucket: str) -> BucketLock:
        if not (lock := self.buckets.get(bucket)):
            lock = LocalBucketLock()
            self.buckets[bucket] = lock

        await self._global.wait()
        return lock

    async def lock_globally(self, release_after: float) -> None:
        create_task(self._lock_global(release_after))



from __future__ import annotations

from asyncio import Task, create_task, sleep
from typing import Awaitable, Callable, Optional, Type

from bauxite.http import HTTPClient, Route

from .enums import EventDirection
from .errors import GatewayCriticalError
from .ratelimiting import GatewayRateLimiter, LocalGatewayRateLimiter
from .shard import Shard, ShardStatusHook

DispatchCallback = Callable[[Shard, EventDirection, dict], Awaitable[None]]


class GatewayClient:
    def __init__(
        self,
        http: HTTPClient,
        intents: int,
        shard_count: Optional[int] = None,
        shard_ids: Optional[list[int]] = None,
        start_limiter: Optional[Type[GatewayRateLimiter]] = None,
        status_hooks: Optional[list[ShardStatusHook]] = None,
        callbacks: Optional[list[DispatchCallback]] = None,
    ) -> None:
        self._http = http

        self._intents = intents

        self._shard_count = shard_count
        self._shard_ids = shard_ids
        self._shard_hooks = status_hooks or []

        self._dispatch_callbacks = callbacks or []

        self._shards: dict[int, Shard] = {}
        self._tasks: dict[int, Task] = {}

        self._gateway: Optional[dict] = None

        self._limiter_class: Type[GatewayRateLimiter] = (
            start_limiter or LocalGatewayRateLimiter
        )
        self._panic: Optional[int] = None

    async def spawn_shards(self) -> None:
        if self._shard_count and not self._shard_ids:
            self._shard_ids = list(range(self._shard_count))

        self._gateway = gateway = await (
            await self._http.request(Route("GET", "/gateway/bot"))
        ).json()

        if self._shard_count:
            assert self._shard_ids

            for id in self._shard_ids:
                self._shards[id] = Shard(
                    id,
                    self._shard_count,
                    self._http._token,
                    self._intents,
                    self._dispatch,
                    self._shard_hooks,
                )
        else:
            for id in range(gateway["shards"]):
                self._shards[id] = Shard(
                    id,
                    gateway["shards"],
                    self._http._token,
                    self._intents,
                    self._dispatch,
                    self._shard_hooks,
                )

        await self._start_shards()

    async def _start_shards(self) -> None:
        assert self._gateway, "Client gateway is not set while starting shards."

        limiter = self._limiter_class(
            self._gateway["session_start_limit"]["max_concurrency"], 5
        )

        for shard in self._shards.values():
            if self._panic is not None:
                raise GatewayCriticalError(self._panic)

            await limiter.wait()

            self._tasks[shard.id] = create_task(self._run_shard(shard))

        while True:
            await sleep(1)

    async def _run_shard(self, shard: Shard) -> None:
        assert (
            self._gateway
        ), f"Client gateway is not set while running shard {shard.id}."

        try:
            await shard.connect(self._http._session, self._gateway["url"])
        except GatewayCriticalError:
            self._panic = True

    async def _dispatch(
        self, shard: Shard, direction: EventDirection, data: dict
    ) -> None:
        for callback in self._dispatch_callbacks:
            await callback(shard, direction, data)

    def get_shard(self, id: int) -> Shard:
        if id not in self._shards:
            raise ValueError(f"Shard of id {id} does not exist.")

        return self._shards[id]



from aiohttp import ClientResponse

from bauxite.error import BauxiteError


class HTTPError(BauxiteError):
    def __init__(self, response: ClientResponse, *args) -> None:
        self.response = response
        self.status = response.status

        super().__init__(*args)


class BadRequest(HTTPError):
    pass


class Unauthorized(BadRequest):
    pass


class Forbidden(BadRequest):
    pass


class NotFound(BadRequest):
    pass


class MethodNotAllowed(BadRequest):
    pass


class UnprocessableEntity(BadRequest):
    pass


class TooManyRequests(BadRequest):
    pass


class ServerError(HTTPError):
    pass


class BadGateway(ServerError):
    pass


class ServiceUnavailable(ServerError):
    pass


class GatewayTimeout(ServerError):
    pass



from io import IOBase
from os import PathLike, path
from typing import Optional, Union


class File:
    """A File object for use when sending messages."""

    __slots__ = ("fp", "_handle", "filename", "_original_close")

    def __init__(
        self,
        fp: Union[IOBase, PathLike, str],
        filename: Optional[str] = None,
        spoiler: bool = False,
    ) -> None:
        if isinstance(fp, IOBase):
            if not (fp.seekable() and fp.readable()):
                raise ValueError(f"IOBase object {fp!r} must be seekable and readable.")

            self.fp = fp
            self._handle = True

        else:
            self.fp = open(fp, "rb")
            self._handle = False

        if filename is None:
            if isinstance(fp, str):
                self.filename: str = path.split(fp)[1]

            else:
                if not hasattr(fp, "name"):
                    if self._handle:
                        self.fp.close()

                    raise ValueError(
                        f"Object given for fp parameter has no name attribute."
                    )

                self.filename: str = fp.name  # type: ignore[attr-defined]
        else:
            self.filename = filename

        if spoiler and not self.filename.startswith("SPOILER_"):
            self.filename = f"SPOILER_{self.filename}"

        self._original_close = (
            self.fp.close
        )  # see: https://github.com/Rapptz/discord.py/blob/master/discord/file.py#L92-L95
        self.fp.close = lambda: None

    def close(self) -> None:
        self.fp.close = self._original_close
        if self._handle:
            self.fp.close()

    def reset(self, hard: Union[bool, int] = True) -> None:
        if hard:
            self.fp.seek(0)



from enum import Enum, IntEnum, auto


class ShardStatus(Enum):
    CONNECTING = auto()
    CONNECTED = auto()
    RESUMING = auto()
    ERRORED = auto()


class EventDirection(Enum):
    OUTBOUND = auto()
    INBOUND = auto()


class GatewayOps(IntEnum):
    DISPATCH = 0
    HEARTBEAT = 1
    IDENTIFY = 2
    PRESENCE_UPDATE = 3
    VOICE_STATE_UPDATE = 4
    RESUME = 6
    RECONNECT = 7
    REQUEST_GUILD_MEMBERS = 8
    INVALID_SESSION = 9
    HELLO = 10
    ACK = 11


class GatewayCloseCodes(IntEnum):
    UNKNOWN_ERROR = 4000
    UNKNOWN_OPCODE = 4001
    DECODE_ERROR = 4002
    NOT_AUTHENTICATED = 4003
    AUTHENTICATION_FAILED = 4004
    ALREADY_AUTHENTICATED = 4005
    INVALID_SEQ = 4007
    RATE_LIMITED = 4008
    SESSION_TIMEOUT = 4009
    INVALID_SHARD = 4010
    SHARDING_REQUIRED = 4011
    INVALID_API_VERSION = 4012
    INVALID_INTENTS = 4013
    DISALLOWED_INTENTS = 4014



from typing import Literal, Optional

HTTPMethod = Literal["GET", "HEAD", "POST", "DELETE", "PUT", "PATCH"]


class Route:
    def __init__(self, method: HTTPMethod, path: str, **params) -> None:
        self.guild_id: Optional[int] = params.get("guild_id")
        self.channel_id: Optional[int] = params.get("channel_id")
        self.webhook_id: Optional[int] = params.get("webhook_id")
        self.webhook_token: Optional[str] = params.get("webhook_token")

        self.method = method
        self.path = path.format(**params)

        self._webhook_bucket: Optional[str] = None
        if self.webhook_id:
            self._webhook_bucket = f"{self.webhook_id}:{self.webhook_token}"

        self.bucket = (
            f"{self.path}-{self.guild_id}:{self.channel_id}:{self._webhook_bucket}"
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bauxite', 'bauxite.gateway', 'bauxite.http']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0']

setup_kwargs = {
    'name': 'bauxite',
    'version': '1.0.0',
    'description': 'A robust, low-level connector for the Discord API',
    'long_description': '# Bauxite\n\nBauxite is a robus, low-level connector for the Discord API.\n\n## What is Bauxite for?\n\nBauxite is made for two main purposes:\n\n- Creating higher-level API wrappers and frameworks\n- Creating things that need high levels of control and low-level access to the Discord API\n\n## Examples\n\n### Basic HTTP Example\n\n```py\nfrom asyncio import run\n\nfrom bauxite import HTTPClient, Route\n\n\nasync def main() -> None:\n    client = HTTPClient("your_bot_token")\n\n    await client.request(\n        Route("POST", "/channels/{channel_id}/messages", channel_id=1234),\n        json={\n            "content": "Hello, world!",\n        },\n    )\n\n    await client.close()\n\nrun(main())\n```\n\n### Basic Gateway Example\n\n```py\nfrom asyncio import run\n\nfrom bauxite import GatewayClient, HTTPClient\n\n\nasync def callback(shard, direction, data) -> None:\n    print(f"{shard} [{direction}]: {data[\'op\'] or data[\'t\']}")\n\nasync def main() -> None:\n    client = HTTPClient("your_bot_token")\n    gateway = GatewayClient(client, 32767, callbacks=[callback])\n\n    await gateway.spawn_shards()\n\nrun(main())\n```\n',
    'author': 'vcokltfre',
    'author_email': 'vcokltfre@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vcokltfre/bauxite',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from bauxite.error import BauxiteError


class GatewayReconnect(BauxiteError):
    pass


class GatewayCriticalError(BauxiteError):
    def __init__(self, code: int) -> None:
        self.code = code

        super().__init__(f"Gateway disconnected with close code {code}")
